/*
								+----------------------------------+
								|                                  |
								| ***  GPU image-proc library  *** |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|            GPU_Lib.h             |
								|                                  |
								+----------------------------------+
*/

#ifndef __GPU_LIB_INCLUDED
#define __GPU_LIB_INCLUDED

/* -config-section- */

/** if defined, DigiLib's <image.h> is included. otherwise local header copy is used. */
//#define GPU_LIB_INCLUDE_DIGILIB

/** if defined, everytime shader is compiled, message is written to stdout. */
//#define GPU_LIB_REPORT_SHADER_COMPILE

/**
 * if defined, everytime shader compilation produces some errors
 * or warnings, they're written to stderr (otherwise silently ignored).
 * it's recommended so you use this first time to use this library on new hardware
 * for case your OpenGL implementation would refuse to compile the shaders.
 */
#define GPU_LIB_REPORT_SHADER_COMPILE_ERRORS

/**
 * if defined, Init_OpenGL() creates small window with OpenGL 2.0 logo, otherwise
 * window remains hidden. works only on windows, linux implementation depends on glut behavior.
 */
#define GPU_LIB_WIN32_SHOW_WINDOW

/** string name of OpenGL logo window. (see GPU_LIB_WIN32_SHOW_WINDOW) */
#define GPU_LIB_OPENGL_WINDOW_NAME "OpenGL"

/**
 * if defined, RGB color format is handled as BGR. BGR is native format for GPU's and
 * therefore any transfers from or to GPU are faster in BGR. histogram and minmax functions
 * swaps values correctly, the only different function is ImageColorMatrix.
 */
#define GPU_LIB_BGR

/**
 * if defined, float value textures will be stored in half (16-bit) floating point format
 * otherwise 'classic' IEEE-754 32-bit floating point format is used. purpose of this is the fact
 * that on older cards, 32-bit fp formats are not accelerated and thus are incredibly slow.
 */
#define GPU_USE_HALF_FLOAT

/**
 * if defined, all float calculations will be done on CPU. this can be useful in case
 * 32-bit fp is not accelerated and 16-bit fp is not precise enough.
 */
//#define GPU_AVOID_FLOAT_COMPUTING // todo

/**
 * use fullscreen quad instead of fullscreen triangle for shader invokation. some believe
 * this is theoretically faster because triangle only needs three vertices. on the other hand,
 * there's extra clipping work. performance gain would supposedly depend on graphics card model.
 */
#define GPU_LIB_USE_FULLSCREEN_TRIANGLE

/* -config-end- */

/* -digilib-defs-begin- */

#ifdef GPU_LIB_INCLUDE_DIGILIB
#include <image.h>
#else //GPU_LIB_INCLUDE_DIGILIB

typedef unsigned char Image8Value;
typedef short Image16Value;
typedef long Image32Value;
typedef float ImageFloatValue;
typedef unsigned char ImageRGBValue;
typedef unsigned long ImageRGBValueRGB;
typedef struct { float Re; float Im; } ImageComplexValue;

typedef struct ImageStruct {
  unsigned char * Raster;
  char * Info;
  void * BinInfo;
  int BinInfoSize;
  void *ExtraInfo;
  int XSize;
  int YSize;
  int XOffset;
  int YOffset;
  int InfoSize;
  short PixelType;
  short ImageType;
  int References;
  union {
    struct {
      void (* Free)(void *);
    } Regular;
    struct {
      void (* Free)(void *);
      struct ImageStruct * Referenced;
    } Reference;
    struct {
      void (* Free)(void *);
      void (* FreeData)(void *);
      void * Data;
    } External;
  } Optional;
} ImageStruct;

/* Pixel Type Definitions */

#define Image8                 ((short)(0x0010|sizeof(Image8Value)))
#define Image8Linear           ((short)(0x0030|sizeof(Image8Value)))
#define Image16                ((short)(0x0040|sizeof(Image16Value)))
#define Image16Linear          ((short)(0x00C0|sizeof(Image16Value)))
#define Image32                ((short)(0x0100|sizeof(Image32Value)))
#define Image32Linear          ((short)(0x0300|sizeof(Image32Value)))
#define ImageFloat             ((short)(0x0400|sizeof(ImageFloatValue)))
#define ImageFloatLinear       ((short)(0x0C00|sizeof(ImageFloatValue)))
#define ImageRGB               ((short)(0x1000|sizeof(ImageRGBValueRGB)))
#define ImageRGBLinear         ((short)(0x3000|sizeof(ImageRGBValueRGB)))
#define ImageComplex           ((short)(0x4000|sizeof(ImageComplexValue)))
#define ImageComplexLinear     ((short)(0xC000|sizeof(ImageComplexValue)))

#define ImageLinearMask        ((short)(0xAAA0))
#define ImageValueMask         0xF

/* Image Type Definitions */

#define RegularImage           0x01
#define ReferenceImage         0x02
#define ExternalImage          0x04
#define TempImage              0x08

/* Info Definitions */

#define DefaultInfoSize        768

#endif //GPU_LIB_INCLUDE_DIGILIB

/* -digilibs-defs-end- */

#if !defined(__cplusplus) && !defined(bool)
#include <stdbool.h>
#endif
// booleans

#ifdef __cplusplus
extern "C" {
#endif

/**
 *	bool Init_OpenGL()
 *		- initializes OpenGL. under linux uses GLUT, under windows uses custom driver
 *		- in case running under windows, there's option to display OpenGL window,
 *		  containing "OpenGL 2.0" logo (enabled using GPU_LIB_WIN32_SHOW_WINDOW #define)
 *		- returns true on success (or multiple initialization), false on failure
 */
bool Init_OpenGL();

/**
 *	bool Free_OpenGL_Objects()
 *		- de-allocates all OpenGL objects. use only in case other than built-in OpenGL
 *		  initialization code was used and therefore it's not possible to use Shutdown_OpenGL()
 *		  for shutdown
 *		- returns true on success, false in case OpenGL was not running
 */
bool Free_OpenGL_Objects();

/**
 *	bool Shutdown_OpenGL();
 *		- shuts down OpenGL, previously initialized by Init_OpenGL(), calls Free_OpenGL_Objects()
 *		  to delete all kinds of temporary objects library uses
 *		- returns true on success, false on failure
 *		- note for linux - there's no function to shut down GLUT, it's expected
 *		  to call this just before exitting as it has really no effect on OpenGL, it frees
 *		  OpenGL objects only
 */
bool Shutdown_OpenGL();

/**
 *	bool b_NPOTTexturesSupported()
 *		- returns true in case non-power-of-two textures are supported, otherwise false
 *		- presumably returns false in case OpenGL is not running
 *		- note extension name is GL_ARB_texture_non_power_of_two (promoted to core in OpenGL 2.0)
 */
bool b_NPOTTexturesSupported();

/**
 *	bool b_FloatTexturesSupported()
 *		- returns true in case float texture pixel-format is supported, otherwise false
 *		- presumably returns false in case OpenGL is not running
 *		- note extension name is GL_ARB_texture_float
 */
bool b_FloatTexturesSupported();

/**
 *	bool b_FramebufferObjectSupported()
 *		- returns true in case frame-buffer objects (FBO's) are supported, otherwise false
 *		- presumably returns false in case OpenGL is not running
 *		- note extension name is GL_EXT_framebuffer_object
 *		- in case FBO's aren't supported, the library is forced to work in "software" mode
 */
bool b_FramebufferObjectSupported();

/**
 *	bool b_ProgrammableShadingSupported()
 *		- returns true in case OpenGL shading language is supported, otherwise false
 *		- presumably returns false in case OpenGL is not running
 *		- note extension name is GL_ARB_shading_language_100, which is core in OpenGL 2.0
 *		- in case shaders aren't supported, the library is forced to work in "software" mode
 */
bool b_ProgrammableShadingSupported();

/**
 *	bool b_ImagingSubsetSupported()
 *		- returns true in case OpenGL's imaging subset and pixel-buffer objects
 *		  are supported, otherwise false
 *		- presumably returns false in case OpenGL is not running
 *		- note extension names are GL_ARB_vertex_buffer_object (promoted to core feature in
 *		  OpenGL 1.5), GL_ARB_pixel_buffer_object (core feature in OpenGL 2.1) and
 *		  GL_ARB_imaging (core feature in OpenGL 1.2)
 *		- GPU accelerated histogram and minmax relies on those and can't work without them
 */
bool b_ImagingSubsetSupported();

/**
 *	bool b_AsyncTransfersSupported()
 *		- returns true in case asynchronous data transfers are supported, otherwise false
 *		- presumably returns false in case OpenGL is not running
 *		- note extension names are GL_ARB_vertex_buffer_object (promoted to core feature in
 *		  OpenGL 1.5) and GL_ARB_pixel_buffer_object (core feature in OpenGL 2.1)
 */
bool b_AsyncTransfersSupported();

/**
 *	bool b_NVFenceSupported()
 *		- returns true in case fence mechanism is supported, otherwise false
 *		- presumably returns false in case OpenGL is not running
 *		- fence mechanism is useful for determining wheter certain OpenGL operations had
 *		  completed, here for querying asynchronous image transfers completeness
 *		- note extension name is GL_NV_fence (not supported by ATI yet)
 *		- without this extension, async transfers lose their sense
 */
bool b_NVFenceSupported();

/**
 *	bool b_GL_ImagesSupported()
 *		- returns true in case OpenGL buffers with contents mapped to system memory are
 *		  supported, otherwise false
 *		- presumably returns false in case OpenGL is not running
 *		- note extension names are GL_ARB_vertex_buffer_object (promoted to core feature in
 *		  OpenGL 1.5) and GL_ARB_pixel_buffer_object (core feature in OpenGL 2.1)
 */
bool b_GL_ImagesSupported();

/**
 *	int n_MaxTextureSize()
 *		- returns maximal 2D texture dimensions handled by current OpenGL implementation
 *		- return value is only valid in case OpenGL was successfuly initialized
 */
int n_MaxTextureSize();

/**
 *	bool b_FastTransfer(const ImageStruct *p_image)
 *		- determines whether p_image could be transfered using a single block copy command
 *		  or wheter it needs slower per-scanline copy loop
 *		- returns true in case p_image can be transfered the fast path, otherwise false
 *		- note all images with interleaved pixels and/or scanlines will have to be transferd
 *		  using the slow path
 */
bool b_FastTransfer(const ImageStruct *p_image);

/**
 *	bool b_HasTexture(const ImageStruct *p_image)
 *		- returns true in case p_image has got associated texture on GPU
 *		- always return false in case OpenGL is not running or was not initialized properly
 *		- note in case image doesn't have texture associated, it shall get one in first
 *		  image processing function where it's subjected either as a parameter or destination
 */
bool b_HasTexture(const ImageStruct *p_image);

/**
 *	bool b_IsGL_Image(const ImageStruct *p_image)
 *		- returns true in case p_image is image in GPU memory mapped to system memory space,
 *		  otherwise returns false (which means reading/writing from/to image raster could mean
 *		  reading/writing directly from/to GPU memory)
 *		- note address of memory mapping could be changed by operations working with this
 *		  image. refer to b_GL_Image_RelocTest().
 *		- always returns false in case p_image has no associated texture
 */
bool b_IsGL_Image(const ImageStruct *p_image);

/**
 *	bool Upload_ImageStruct(ImageStruct *p_image)
 *		- forces upload of p_image from system memory to GPU memory
 *		- returns true on success, false on failure
 */
bool Upload_ImageStruct(ImageStruct *p_image);

/**
 *	bool Download_ImageStruct(ImageStruct *p_image)
 *		- forces download of p_image from GPU memory to system memory
 *		- returns true on success, false on failure
 */
bool Download_ImageStruct(ImageStruct *p_image);

/**
 *	bool Upload_Async_ImageStruct(ImageStruct *p_image, unsigned int *p_fence_id)
 *		- starts asynchronous upload of p_image from system memory to GPU memory
 *		  (returns almost immediately)
 *		- p_fence_id is output parameter, it will be written task identifier which can
 *		  be used in b_FenceFinished() and WaitFence() functions
 *		- in case fences aren't supported (refer to b_NVFenceSupported()), contents
 *		  of p_fence_id are undefined
 *		- returns true on success, false on failure
 */
bool Upload_Async_ImageStruct(ImageStruct *p_image, unsigned int *p_fence_id);

/**
 *	bool Download_Async_ImageStruct(ImageStruct *p_image, unsigned int *p_fence_id)
 *		- starts asynchronous download of p_image from GPU memory to system memory
 *		  (returns almost immediately)
 *		- p_fence_id is output parameter, it will be written task identifier which can
 *		  be used in b_FenceFinished() and WaitFence() functions
 *		- in case fences aren't supported (refer to b_NVFenceSupported()), contents
 *		  of p_fence_id are undefined
 *		- returns true on success, false on failure
 */
bool Download_Async_ImageStruct(ImageStruct *p_image, unsigned int *p_fence_id);

/**
 *	bool b_FenceFinished(unsigned int n_fence_id)
 *		- queries state of task with identifier n_fence_id (returns immediately)
 *		- returns true in case task has been already finished, otherwise false
 *		- always returns false in case fences aren't supported (refer to b_NVFenceSupported())
 */
bool b_FenceFinished(unsigned int n_fence_id);

/**
 *	bool WaitFence(unsigned int n_fence_id)
 *		- waits until task with identifier n_fence_id is finished
 *		- returns false in case n_fence_id doesn't contain valid task id, otherwise true
 *		- in case fences aren't supported (refer to b_NVFenceSupported()), waits until
 *		  all pending OpenGL operations are finished
 */
bool WaitFence(unsigned int n_fence_id);

/**
 *	void Enable_AutoUploads()
 *		- enables automatic image uploads, meaning image is copied form system memory
 *		  to GPU memory everytime it's used as source parameter of some image processing function
 *		- note this is default behavior, for compatibility purposes
 */
void Enable_AutoUploads();

/**
 *	void Enable_AutoDownloads()
 *		- enables automatic image downloads, meaning image is copied form GPU memory
 *		  to system memory everytime it's used as destination of some image processing function
 *		- note this is default behavior, for compatibility purposes
 */
void Enable_AutoDownloads();

/**
 *	void Disable_AutoUploads()
 *		- disables automatic image uploads, meaning image is copied form system memory
 *		  to GPU memory when it's used as source parameter of some image processing function
 *		  and it's dirty flag is set to false (ie. current on client)
 *		- doesn't affect forced image transfers
 */
void Disable_AutoUploads();

/**
 *	void Disable_AutoDownloads()
 *		- disables automatic image downloads, meaning image is not copied form GPU memory
 *		  to system memory after used as destination of some image processing function
 *		- doesn't affect forced image transfers
 *		- note in case automatic downloads are disabled and automatic uploads enabled,
 *		  image that is rendered to and then again used in rendering will first not be
 *		  copied to it's imagestruct and then will be rewritten by old data from it's imagestruct
 */
void Disable_AutoDownloads();

/**
 *	bool b_AutoDownloads_Enabled()
 *		- returns true in case automatic downloads are enabled, otherwise false
 */
bool b_AutoDownloads_Enabled();

/**
 *	bool b_AutoUploads_Enabled()
 *		- returns true in case automatic uploads are enabled, otherwise false
 */
bool b_AutoUploads_Enabled();

/**
 *	bool Set_DirtyFlag(const ImageStruct *p_image, bool b_current_on_server)
 *		- used to set image transfer goals in case automatic transfers are disabled
 *		- sets image dirty flag. in case b_current_on_server is true, copy of image in GPU
 *		  memory is considered valid. in case b_current_on_server is false, original image
 *		  in system memory is considered valid and shall be uploaded if needed even though
 *		  automatic uploads might be disabled
 *		- in case goal was to invalidate GPU-side copy and image has no associated texture,
 *		  returns false. otherwise returns true (in case image has no texture, it would have
 *		  to be copied anyway if needed as source to some image processing function)
 *		- doesn't affect forced image transfers
 */
bool Set_DirtyFlag(const ImageStruct *p_image, bool b_current_on_server);

/**
 *	bool b_GL_Image_RelocTest()
 *		- runs a simple test to find out wheter mapped buffers changes their address
 *		  in system memory while being mapped and unmapped
 *		- returns false in case it is not possible to rely on constant buffer address
 *		  (meaning constant mapped image Raster pointer), otherwise true
 *		- note this is only a test and therefore cannot be fully relied on
 */
bool b_GL_Image_RelocTest();

/**
 *	ImageStruct *p_Create_GL_Image(int n_width, int n_height, short n_format)
 *		- allocates an empty image. image has width n_width, height n_height
 *		  and pixel format n_format
 *		- returns pointer to a new image or NULL in case there was not enough memory
 *		- it's Optional.Regular.Free pointer is set to function which would properly free the image
 */
ImageStruct *p_Alloc_Image(int n_width, int n_height, short n_format);

/**
 *	ImageStruct *p_Create_GL_Image(int n_width, int n_height, short n_format)
 *		- creates image with mapped pixel buffer. image has width n_width, height n_height
 *		  and pixel format n_format
 *		- returns pointer to a new image or NULL in case there was not enough memory or in case
 *		  mapped images aren't supported
 *		- note the Raster pointer of such an image might change accross calls to image
 *		  processing functions, using this image. refer to b_GL_Image_RelocTest()
 *		- it's Optional.Regular.Free pointer is set to function which would properly free the image
 */
ImageStruct *p_Create_GL_Image(int n_width, int n_height, short n_format);

/**
 *	void Delete_GL_Image(void *p_image)
 *		- deletes associated texture of p_image in case it has got one
 *		- in case p_image is image, created using p_Create_GL_Image, it's completely deleted,
 *		  including ImageStruct and it's Raster memory
 */
void Delete_GL_Image(void *p_image);

/**
 *	bool CheckBufferMapping(ImageStruct *p_image)
 *		- in case there was an error, sometimes it's necessary to check buffer mapping and
 *		  maybe renew it
 *		- returns true in case buffer data was not changed (though image data are in texture
 *		  but they might be out-of-date in case user changed image Raster contents),
 *		  returns false on failure or in case data was lost
 *		- note it's necessary to check p_image->Raster as it might be NULL if function fails
 */
bool CheckBufferMapping(ImageStruct *p_image);

/**
 *	bool b_OpenGLError()
 *		- returns true in case OpenGL error(s) occured, otherwise false
 *		- note this clears the error flag so it wouldn't return true untill some more
 *		  OpenGL errors occured
 *		- note this calls glGetError() only
 */
bool b_OpenGLError();

/**
 *	unsigned int n_GPU_ImageStruct_Num()
 *		- returns number of ImageStruct copies, residing on GPU
 */
unsigned int n_GPU_ImageStruct_Num();

/**
 *	unsigned int n_Shader_Num()
 *		- returns number of compiled shaders
 */
unsigned int n_Shader_Num();

/**
 *	unsigned int n_RenderBuffer_Num()
 *		- returns number of render-buffers
 */
unsigned int n_RenderBuffer_Num();

/**
 *	unsigned int n_TempTexture_Num()
 *		- returns number of temporary textures
 */
unsigned int n_TempTexture_Num();

/**
 *	unsigned int n_TempTexture_Size()
 *		- returns amount of memory taken by temporary textures
 *		- note there are no functions for querying size of shaders nor render-buffers
 *		  because they take up virtually no memory (render-buffers are just contexts,
 *		  not real buffers)
 */
unsigned int n_TempTexture_Size();

/**
 *	void Purge_TempTextures_Size(unsigned int n_size)
 *		- free at least n_size bytes of space from temporary textures
 *		- this may come in handy if there is far too much temporary textures
 *		  and processing is slow due to GPU memory swapping
 *		- note temporary textures accumulate with different image formats and sizes
 *		  it was necessary to work with
 */
void Purge_TempTextures_Size(unsigned int n_size);

/**
 *	bool ImageCopy(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- copies source image to dest image. converts pixel types and
 *		  has free bilinear resampling, same as all the others
 *		- returns true on success, false on failure
 */
bool ImageCopy(ImageStruct *p_dest_image, ImageStruct *p_src_image);

/**
 *	bool ImageImageAdd(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
 *		- image addition point operator
 */
bool ImageImageAdd(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1);

/**
 *	bool ImageImageSub(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
 *		- image subtraction point operator
 */
bool ImageImageSub(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1);

/**
 *	bool ImageImageMult(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
 *		- image multiplication point operator
 */
bool ImageImageMult(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1);

/**
 *	bool ImageImageDiv(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
 *		- image division point operator
 */
bool ImageImageDiv(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1);

/**
 *	bool ImageImageBlend(ImageStruct *p_dest_image, ImageStruct *p_src_image0,
 *		ImageStruct *p_src_image1, float f_weight0, float f_weight1)
 *		- calculates following equation: p_src_image0 * f_weight0 + p_src_image1 * f_weight1
 *		- weight magnitudes are unlimited, results are clamped based on dest image pixelformat
 */
bool ImageImageBlend(ImageStruct *p_dest_image, ImageStruct *p_src_image0,
	ImageStruct *p_src_image1, float f_weight0, float f_weight1);

/**
 *	bool ImageImageDecal(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
 *		- alpha-blends source image 1 onto source image 0, ie:
 *		  p_src_image0 * (1 - p_src_image1_alpha) + p_src_image1 * p_src_image1_alpha
 */
bool ImageImageDecal(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1);

/**
 *	bool ImageInvert(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- calculates pointwise 1 - p_src_image
 */
bool ImageInvert(ImageStruct *p_dest_image, ImageStruct *p_src_image);

/**
 *	bool ImageAbs(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- calculates pointwise absolute value of p_src_image
 */
bool ImageAbs(ImageStruct *p_dest_image, ImageStruct *p_src_image);

/**
 *	bool ImageImageMin(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
 *		- pointwise minimum of p_src_image0 and p_src_image1
 */
bool ImageImageMin(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1);

/**
 *	bool ImageImageMax(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
 *		- pointwise maximum of p_src_image0 and p_src_image1
 */
bool ImageImageMax(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1);

/**
 *	bool ImageThresh(ImageStruct *p_dest_image, ImageStruct *p_src_image, float f_threshold)
 *		- thresholds image to f_threshold
 *		- note threshold value is relative to image pixel-type value range, for example
 *		  threshold 0.5 corresponds with both 128 for Image8 and 32767 for Image16
 *		- floating-point images are compared without any value rescaling
 */
bool ImageThresh(ImageStruct *p_dest_image, ImageStruct *p_src_image, float f_threshold);

/**
 *	bool ImageLog(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- calculates pointwise log(1 + p_src_image)
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 */
bool ImageLog(ImageStruct *p_dest_image, ImageStruct *p_src_image);

/**
 *	bool ImageExp(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- calculates pointwise exp(p_src_image) - 1
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 */
bool ImageExp(ImageStruct *p_dest_image, ImageStruct *p_src_image);

/**
 *	bool ImageSqrt(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- calculates pointwise square root of p_src_image
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 *		- note this is somewhat slower than ImageLog
 */
bool ImageSqrt(ImageStruct *p_dest_image, ImageStruct *p_src_image);

/**
 *	bool ImagePow(ImageStruct *p_dest_image, ImageStruct *p_src_image, float f_power)
 *		- pointwise raises p_src_image to f_power
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 *		- note this is somewhat slower than ImageExp
 */
bool ImagePow(ImageStruct *p_dest_image, ImageStruct *p_src_image, float f_power);

/**
 *	bool ImageScale(ImageStruct *p_dest_image, ImageStruct *p_src_image, float f_scale, float f_bias)
 *		- pointwise image scaling as p_src_image * f_scale + f_bias
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 *		- scale and bias values are unlimited
 */
bool ImageScale(ImageStruct *p_dest_image, ImageStruct *p_src_image, float f_scale, float f_bias);

/**
 *	bool ImageColorMatrix(ImageStruct *p_dest_image, ImageStruct *p_src_image,
 *		const float *p_mat4x4, bool b_transpose)
 *		- treats each point of p_src_image as four-dimensional vector (RGBA),
 *		  multiplies it with matrix p_mat4x4 which can be applied as is or transposed
 *		- note fixed-point pixel values are rescaled to 0 - 1 range
 *		- matrix elements values are unlimited
 *		- for example this matrix can be used to convert RGB images to YCbCr colorspace:
 *		  (note alpha channel of such an image must contain value 1 (respectively 255))
 *
 *		float p_mat4x4[16] = { // RGB to YCbCr (ITU-R BT.601)
 *			    .299f,     .587f,     .114f,   0, // R' = .299R + .587G + .114B
 *			-.168736f, -.331264f,       .5f, .5f, // G' = .5 - .168736R - .331264G - .5B
 *			      .5f, -.418688f, -.081312f, .5f, // B' = .5 + .5R - .418688G - .081312B
 *			        0,         0,         0,   1  // A' = 1
 *		};
 */
bool ImageColorMatrix(ImageStruct *p_dest_image, ImageStruct *p_src_image,
	const float *p_mat4x4, bool b_transpose);

/**
 *	bool ImageGreyscale(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- turns source image into greyscale, using std ".299, .587, .144" equation
 *		- in case dest is RGBA image, RGB are set to gray value, A is set to 1
 */
bool ImageGreyscale(ImageStruct *p_dest_image, ImageStruct *p_src_image);

/**
 *	bool ImageResize(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- resamples image p_src_image to p_dest_image
 *		- uses bicubic filter
 */
bool ImageResize(ImageStruct *p_dest_image, ImageStruct *p_src_image);

/**
 *	bool ImageRotate(ImageStruct *p_dest_image, ImageStruct *p_src_image)
 *		- rotates p_src_image by f_angle in radians, clockwise
 *		- image proportions are kept intact (in contrast from ImageResize which don't mind them)
 *		- in case b_scaledown_to_fit is true, source image is scaled down so it's whole area
 *		  is visible (otherwise, rotated corners might be reaching out)
 *		- f_scale is additional scaling ratio (noscale is 1.0)
 *		- uses bicubic filter
 */
bool ImageRotate(ImageStruct *p_dest_image, ImageStruct *p_src_image,
	float f_angle, bool b_scaledown_to_fit, float f_scale);

/**
 *	bool ImageTransform(ImageStruct *p_dest_image, ImageStruct *p_src_image,
 *		float x0, float y0, float x1, float y1, float x2, float y2, float x3, float y3)
 *		- arbitrary image interpolation over convex quad given as (x0, y0), (x1, y1), ...
 *		- coordinates in OpenGL-ish fashion, orthographic projection
 */
bool ImageTransform(ImageStruct *p_dest_image, ImageStruct *p_src_image,
	float x0, float y0, float x1, float y1, float x2, float y2, float x3, float y3);

/**
 *	bool ImageAffineTransform(ImageStruct *p_dest_image, ImageStruct *p_src_image,
 *		float x0, float y0, float z0, float x1, float y1, float z1,
 *		float x2, float y2, float z2, float x3, float y3, float z3)
 *		- arbitrary image interpolation over convex quad given as (x0, y0, z0), (x1, y1, z1), ...
 *		- coordinates in OpenGL-ish fashion, orthographic projection
 */
bool ImageAffineTransform(ImageStruct *p_dest_image, ImageStruct *p_src_image,
	float x0, float y0, float z0, float x1, float y1, float z1,
	float x2, float y2, float z2, float x3, float y3, float z3);

/**
 *	bool ImageSquareErode(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- erosion morphologic operator
 *		- structuring element is square with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 */
bool ImageSquareErode(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h);

/**
 *	bool ImageSquareDilate(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- dilation morphologic operator
 *		- structuring element is square with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 */
bool ImageSquareDilate(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h);

/**
 *	bool ImageSquareOpen(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- opening morphologic operator
 *		- structuring element is square with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 */
bool ImageSquareOpen(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h);

/**
 *	bool ImageSquareClose(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- closing morphologic operator
 *		- structuring element is square with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 */
bool ImageSquareClose(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h);

/**
 *	bool ImageDiamondErode(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- erosion morphologic operator
 *		- structuring element is diamond with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- note "diamond" shape is chamfered square
 */
bool ImageDiamondErode(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h);

/**
 *	bool ImageDiamondDilate(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- dilation morphologic operator
 *		- structuring element is diamond with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- note "diamond" shape is chamfered square
 */
bool ImageDiamondDilate(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h);

/**
 *	bool ImageDiamondOpen(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- opening morphologic operator
 *		- structuring element is diamond with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- note "diamond" shape is chamfered square
 */
bool ImageDiamondOpen(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h);

/**
 *	bool ImageDiamondClose(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- closing morphologic operator
 *		- structuring element is diamond with dimensions n_kernel_w  n_kernel_h pixels
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- note "diamond" shape is chamfered square
 */
bool ImageDiamondClose(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h);

/**
 *	bool ImageImageErode(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- erosion morphologic operator
 *		- structuring element is p_src_image1
 *		- note source image nor structuring element has to be greyscale, they will be
 *		  thresholded to 0.5 when sampled
 */
bool ImageImageErode(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1);

/**
 *	bool ImageImageDilate(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- dilation morphologic operator
 *		- structuring element is p_src_image1
 *		- note source image nor structuring element has to be greyscale, they will be
 *		  thresholded to 0.5 when sampled
 */
bool ImageImageDilate(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1);

/**
 *	bool ImageImageOpen(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- opening morphologic operator
 *		- structuring element is p_src_image1
 *		- note source image nor structuring element has to be greyscale, they will be
 *		  thresholded to 0.5 when sampled
 */
bool ImageImageOpen(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1);

/**
 *	bool ImageImageClose(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- closing morphologic operator
 *		- structuring element is p_src_image1
 *		- note source image nor structuring element has to be greyscale, they will be
 *		  thresholded to 0.5 when sampled
 */
bool ImageImageClose(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1);

/**
 *	bool ImageImageHitMiss(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1)
 *		- hit-and-miss morphologic operator
 *		- structuring element is p_src_image1
 *		- note source image doesn't have to be greyscale, it will be thresholded to 0.5
 *		- structuring element is thresholded as follows: values below .25 are "zero",
 *		  values above .75 are "one", values between .25 and .75 are "don't care"
 */
bool ImageImageHitMiss(ImageStruct *p_dest_image, ImageStruct *p_src_image0, ImageStruct *p_src_image1);

/**
 *	bool ImageMean(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- separable (two-pass) mean image filter
 *		- averaging window size is n_kernel_w  n_kernel_h pixels
 */
bool ImageMean(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h);

/**
 *	bool ImageMedian(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_size)
 *		- separable (two-pass) median image filter
 *		- averaging window size is n_kernel_size  n_kernel_size pixels, the only supported
 *		  sizes are 3  3, 5  5, 7  7 and 15  15.
 */
bool ImageMedian(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_size);

/**
 *	bool ImageLocalMin(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- finds local minimum under the window
 *		- window size is n_kernel_w  n_kernel_h pixels
 */
bool ImageLocalMin(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h);

/**
 *	bool ImageLocalMax(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h)
 *		- finds local maximum under the window
 *		- window size is n_kernel_w  n_kernel_h pixels
 */
bool ImageLocalMax(ImageStruct *p_dest_image, ImageStruct *p_src_image, int n_kernel_w, int n_kernel_h);

/**
 *	bool ImageConvolveSeparable(ImageStruct *p_dest_image, ImageStruct *p_src_image,
 *		ImageStruct *p_kernel_h, ImageStruct *p_kernel_v)
 *		- separable convolution filter
 *		- note horizontal and vertical kernel doesn't have to be aligned in horizontal
 *		  or vertical direction respectively, they can both point to single image
 *		- note this handles convolution with complex kernels and / or complex image
 */
bool ImageConvolveSeparable(ImageStruct *p_dest_image, ImageStruct *p_src_image,
	ImageStruct *p_kernel_h, ImageStruct *p_kernel_v);

/**
 *	bool ImageConvolve2D(ImageStruct *p_dest_image, ImageStruct *p_src_image, ImageStruct *p_kernel)
 *		- two-dimensional convolution filter
 *		- note this handles convolution with complex kernel and / or complex image
 */
bool ImageConvolve2D(ImageStruct *p_dest_image, ImageStruct *p_src_image, ImageStruct *p_kernel);

/**
 *	bool ImageGetHistogram(ImageStruct *p_src_image, int n_histogram_width, unsigned int *p_histogram)
 *		- retrieves image histogram
 *		- n_histogram_width is histogram width and it must be power of two
 *		- p_histogram will contain histogram data, it doesn't have to be zeroed prior calling this
 */
bool ImageGetHistogram(ImageStruct *p_src_image, int n_histogram_width, unsigned int *p_histogram);

/**
 *	bool ImageGetLuminosityHistogram(ImageStruct *p_src_image,
 *		int n_histogram_width, unsigned int *p_histogram)
 *		- retrieves image luminosity histogram (greyscale conversion applied)
 *		- n_histogram_width is histogram width and it must be power of two
 *		- p_histogram will contain histogram data, it doesn't have to be zeroed prior calling this
 */
bool ImageGetLuminosityHistogram(ImageStruct *p_src_image,
	int n_histogram_width, unsigned int *p_histogram);

/**
 *	bool ImageGetRGBHistogram(ImageStruct *p_src_image, int n_histogram_width,
 *		unsigned int *p_histo_r, unsigned int *p_histo_g, unsigned int *p_histo_b)
 *		- retrieves RGB image histogram (but it doesn't rely on format of input image)
 *		- n_histogram_width is histogram width and it must be power of two
 *		- p_histo_r, p_histo_g and p_histo_b will contain histogram data, they doen't have
 *		  to be zeroed prior calling this
 */
bool ImageGetRGBHistogram(ImageStruct *p_src_image, int n_histogram_width,
	unsigned int *p_histo_r, unsigned int *p_histo_g, unsigned int *p_histo_b);

/**
 *	bool ImageGetRGBAHistogram(ImageStruct *p_src_image, int n_histogram_width,
 *		unsigned int *p_histo_r, unsigned int *p_histo_g,
 *		unsigned int *p_histo_b, unsigned int *p_histo_a)
 *		- retrieves RGBA image histogram (but it doesn't rely on format of input image)
 *		- n_histogram_width is histogram width and it must be power of two
 *		- p_histo_r, p_histo_g, p_histo_b and p_histo_a will contain histogram data, they
 *		  doen't have to be zeroed prior calling this
 */
bool ImageGetRGBAHistogram(ImageStruct *p_src_image, int n_histogram_width,
	unsigned int *p_histo_r, unsigned int *p_histo_g,
	unsigned int *p_histo_b, unsigned int *p_histo_a);

/**
 *	bool ImageGetMinMax(ImageStruct *p_src_image, void *p_minmax)
 *		- get image minimal and maximal values
 *		- values are written to p_minmax in image pixel-type format in order min, then max
 */
bool ImageGetMinMax(ImageStruct *p_src_image, void *p_minmax);

typedef void *FFTHandle;
#define InvalidFFTHandle (0)

/*
 *	FFTHandle h_CreateFFT(int n_width, bool b_forward)
 *		- returns new FFT handle for n_width samples, in case b_forward is true, it will be
 *		  forward FFT, otherwise reverse FFT
 *		- returns InvalidFFTHandle for invalid widths (non power of two, too small)
 */
FFTHandle h_CreateFFT(int n_width, bool b_forward);

/*
 *	void DeleteFFT(FFTHandle h_fft)
 *		- deletes FFT handle h_fft, obtained by h_CreateFFT function
 */
void DeleteFFT(FFTHandle h_fft);

/*
 *	bool ImageFFT(ImageStruct *p_dest_image, ImageStruct *p_src_image,
 *		FFTHandle h_fft_x, FFTHandle h_fft_y)
 *		- calculates fast fourier transform of p_src_image which needs to be monochrome or complex
 *		  (color images are converted to monochrome by sampling red channel only)
 *		- h_fft_x and h_fft_y are FFT handles for horizontal and vertical pass respectively
 *		  (their size needs to match source image width and height)
 *		- source and dest image sizes must match
 */
bool ImageFFT(ImageStruct *p_dest_image, ImageStruct *p_src_image,
	FFTHandle h_fft_x, FFTHandle h_fft_y);

#ifdef __cplusplus
}
#endif

#endif //__GPU_LIB_INCLUDED
