/*
								+----------------------------------+
								|                                  |
								| *** Image transfer functions *** |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|        ImageTransfer.cpp         |
								|                                  |
								+----------------------------------+
*/

#include <vector>
#include "../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h>
#endif
#define GL_GLEXT_LEGACY
#include <GL/gl.h>
#include "../berLame_src/OpenGL20.h"
#include "../berLame_src/OpenGLState.h"
#include "../berLame_src/Texture.h"
#include "GPU_Lib.h"
#include "ImageTransfer.h"

/**
 *	unsigned int n_PixelSize(int n_pixel_type)
 *		- returns size of DigiLib pixel type n_pixel_type
 */
unsigned int n_PixelSize(int n_pixel_type)
{
	return (unsigned)n_pixel_type & ImageValueMask;
	/*switch(n_pixel_type) {
	case Image8:
	case Image8Linear:
		return sizeof(unsigned char);
	case Image16:
	case Image16Linear:
		return sizeof(short);
	case Image32:
	case Image32Linear:
		return sizeof(long);
	case ImageFloat:
	case ImageFloatLinear:
		return sizeof(float);
	case ImageRGB:
	case ImageRGBLinear:
		return sizeof(unsigned long);
	case ImageComplex:
	case ImageComplexLinear:
		return 2 * sizeof(float);
	default:
		return 0;
	}*/
}

/**
 *	unsigned int n_GL_PixelSize(GLenum n_internal_format)
 *		- returns size of OpenGL pixel type n_internal_format
 */
unsigned int n_GL_PixelSize(GLenum n_internal_format)
{
	switch(n_internal_format) {
	case GL_LUMINANCE8:
		return 1;
	case GL_LUMINANCE16F_ARB:
	case GL_LUMINANCE16:
		return 2;
	case GL_BGRA:
	case GL_RGBA:
	case GL_LUMINANCE32F_ARB:
	case GL_LUMINANCE_ALPHA16F_ARB:
		return 4;
	case GL_LUMINANCE_ALPHA32F_ARB:
		return 8;
	default:
		return 1;
	}
}

/**
 *	bool Get_GL_PixelType(int n_pixel_type, int &r_n_fmt,
 *		int &r_n_data_type, int &r_n_internal_fmt)
 *		- converts DigiLib pixel type n_pixel_type to OpenGL pixel type
 *		- returns true on success or false in case it couldn't find equivalent format
 */
bool Get_GL_PixelType(int n_pixel_type, int &r_n_fmt, int &r_n_data_type, int &r_n_internal_fmt)
{
	r_n_fmt = GL_LUMINANCE; // 1 channel only, converts to red, 0, 0 when displayed
	r_n_data_type = GL_UNSIGNED_BYTE;

	switch(n_pixel_type) {
	case ImageRGB:
	case ImageRGBLinear:
#ifdef GPU_LIB_BGR
		r_n_fmt = GL_BGRA;
#else //GPU_LIB_BGR
		r_n_fmt = GL_RGBA; // slower transfers
#endif //GPU_LIB_BGR
		r_n_internal_fmt = GL_RGBA;
		break;
	case Image8:
	case Image8Linear:
		//r_n_data_type = GL_UNSIGNED_BYTE;
		r_n_internal_fmt = GL_LUMINANCE8;
		break;
	case Image16:
	case Image16Linear:
		r_n_data_type = GL_SHORT;
		r_n_internal_fmt = GL_LUMINANCE16;
		break;
	case Image32:
	case Image32Linear:
		r_n_data_type = GL_INT;
		r_n_internal_fmt = GL_LUMINANCE16; // beware rounding !!!
		break;
	case ImageFloat:
	case ImageFloatLinear:
		r_n_data_type = GL_FLOAT;
#ifdef GPU_USE_HALF_FLOAT
		r_n_internal_fmt = GL_LUMINANCE16F_ARB; // 32-bit luminance not eccelerated :(
#else //GPU_USE_HALF_FLOAT
		r_n_internal_fmt = GL_LUMINANCE32F_ARB;
#endif //GPU_USE_HALF_FLOAT
		break;
	case ImageComplex:
	case ImageComplexLinear:
		r_n_fmt = GL_LUMINANCE_ALPHA; // two - channel luminance-alpha image
		r_n_data_type = GL_FLOAT;
#ifdef GPU_USE_HALF_FLOAT
		r_n_internal_fmt = GL_LUMINANCE_ALPHA16F_ARB;
#else //GPU_USE_HALF_FLOAT
		r_n_internal_fmt = GL_LUMINANCE_ALPHA32F_ARB;
#endif //GPU_USE_HALF_FLOAT
		break;
	default:
		return false;
	}

	return true;
}

/**
 *	bool CopyImage_To_Texture2D(const ImageStruct *p_image)
 *		- copies image p_image to currently bound texture in active texture unit
 *		- returns true on success, false on failure (bad pixel type)
 */
bool CopyImage_To_Texture2D(const ImageStruct *p_image)
{
	int n_format;
	int n_data_type;
	int n_internal_format;

	if(!Get_GL_PixelType(p_image->PixelType, n_format, n_data_type, n_internal_format))
		return false;

	const void *p_raster = p_image->Raster;
	int n_y_offset = abs(p_image->YOffset);
	int n_x_offset = abs(p_image->XOffset);
	if(p_image->YOffset < 0)
		p_raster = (const void*)((long)p_raster - n_y_offset * (p_image->YSize - 1));
	if(p_image->XOffset < 0)
		p_raster = (const void*)((long)p_raster - n_x_offset * (p_image->XSize - 1));
	// find upper-left corner of image

	int n_pixel_size = n_PixelSize(p_image->PixelType);
	// get pixel size to determine offsets in pixels instead of bytes

	glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
	glPixelStorei(GL_UNPACK_SKIP_IMAGES, 0);
	glPixelStorei(GL_UNPACK_IMAGE_HEIGHT, 0);
	// make sure they are in their default state

	if(n_x_offset == n_pixel_size) {
		if(!(n_y_offset % n_pixel_size)) {
			glPixelStorei(GL_UNPACK_ALIGNMENT, n_x_offset);
			glPixelStorei(GL_UNPACK_ROW_LENGTH, (n_y_offset != n_x_offset * p_image->XSize)?
				n_y_offset / n_pixel_size : 0);

			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, p_image->XSize,
				p_image->YSize, n_format, n_data_type, p_raster);
			// copy image as a single block (fast)

			return true;
		} else {
			glPixelStorei(GL_UNPACK_ALIGNMENT, n_x_offset);
			glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);

			const unsigned char *p_raster_b = (const unsigned char*)p_raster;
			for(int y = 0; y < p_image->YSize; ++ y) {
				glTexSubImage2D(GL_TEXTURE_2D, 0, 0, y, p_image->XSize, 1,
					n_format, n_data_type, p_raster_b + y * n_y_offset);
			}
			// copy image as horizontal scanlines (slow)

			return true;
		}
	} else {
		if(!(n_y_offset % n_pixel_size)) {
			glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
			glPixelStorei(GL_UNPACK_ROW_LENGTH, n_y_offset / n_pixel_size); // !!

			const unsigned char *p_raster_b = (const unsigned char*)p_raster;
			for(int x = 0; x < p_image->XSize; ++ x) {
				glTexSubImage2D(GL_TEXTURE_2D, 0, x, 0, 1, p_image->YSize,
					n_format, n_data_type, p_raster_b + x * n_x_offset);
			}
			// copy image as vertical scanlines (slow, slower than h-scanlines because of cache)

			return true;
		} else {
			glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
			glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);

			const unsigned char *p_raster_b = (const unsigned char*)p_raster;
			for(int y = 0; y < p_image->YSize; ++ y) {
				for(int x = 0; x < p_image->XSize; ++ x) {
					glTexSubImage2D(GL_TEXTURE_2D, 0, x, y, 1, 1,
						n_format, n_data_type, p_raster_b + x * n_x_offset + y * n_y_offset);
				}
			}
			// copy image as pixels (*very* slow!)

			return true;
		}
	}
}

/**
 *	bool CopyTexture2D_To_Image(const ImageStruct *p_image)
 *		- copies image from currently bound texture in active texture unit to p_image
 *		- returns true on success, false on failure (bad pixel type)
 */
bool CopyTexture2D_To_Image(ImageStruct *p_image)
{
	int n_format;
	int n_data_type;
	int n_internal_format;

	if(!Get_GL_PixelType(p_image->PixelType, n_format, n_data_type, n_internal_format))
		return false;

	void *p_raster = p_image->Raster;
	int n_y_offset = abs(p_image->YOffset);
	int n_x_offset = abs(p_image->XOffset);
	if(p_image->YOffset < 0)
		p_raster = (void*)((long)p_raster - n_y_offset * (p_image->YSize - 1));
	if(p_image->XOffset < 0)
		p_raster = (void*)((long)p_raster - n_x_offset * (p_image->XSize - 1));
	// find upper-left corner of image

	int n_pixel_size = n_PixelSize(p_image->PixelType);
	// get pixel size to determine offsets in pixels instead of bytes

	glPixelStorei(GL_PACK_SKIP_ROWS, 0);
	glPixelStorei(GL_PACK_SKIP_PIXELS, 0);
	glPixelStorei(GL_PACK_SKIP_IMAGES, 0);
	glPixelStorei(GL_PACK_IMAGE_HEIGHT, 0);
	// make sure they are in their default state

	if(n_x_offset == n_pixel_size && !(n_y_offset % n_pixel_size)) {
		glPixelStorei(GL_PACK_ALIGNMENT, n_x_offset);
		glPixelStorei(GL_PACK_ROW_LENGTH, (n_y_offset != n_x_offset * p_image->XSize)?
			n_y_offset / n_pixel_size : 0);

		glGetTexImage(GL_TEXTURE_2D, 0, n_format, n_data_type, p_raster);
		// copy image as a single block (fast)

		return true;
	} else {
		if(n_pixel_size < 1 || n_pixel_size > 8)
			return false;

		glPixelStorei(GL_PACK_ALIGNMENT, n_pixel_size);
		glPixelStorei(GL_PACK_ROW_LENGTH, 0);

		glGetTexImage(GL_TEXTURE_2D, 0, n_format, n_data_type, p_raster);
		// copy image as a single block (fast)

		unsigned char *p_raster_b = (unsigned char*)p_raster;
		for(int y = p_image->YSize - 1, n_yo = p_image->YOffset,
		   n_yo2 = n_pixel_size * p_image->XSize; y >= 0; -- y) {
			for(int x = p_image->XSize - 1, n_xo = p_image->XOffset; x >= 0; -- x) {
				const unsigned char *p_src_pixel = p_image->Raster +
					n_pixel_size * x + n_yo2 * y;
				unsigned char *p_dest_pixel = p_image->Raster + n_xo * x + n_yo * y;
				// get pixel address for both source and destination images

				if(n_pixel_size == 1)
					*p_dest_pixel = *p_src_pixel;
				else if(n_pixel_size == 4)
					*(int*)p_dest_pixel = *(int*)p_src_pixel;
				else if(n_pixel_size == 2)
					*(short*)p_dest_pixel = *(short*)p_src_pixel;
				else
					memcpy(p_dest_pixel, p_src_pixel, n_pixel_size);
				// several branches for different pixel sizes, allowing compiler to optimize
			}
		}
		// post-op shift image pixels to required positions

		// there's no glGetTexSubImage so there's no other way than direct block transfer
		// and subsequent pixel interleaving

		return true;
	}
}

/**
 *	CGLTexture_2D *p_Create_ImageStruct_Texture(CGLState *p_state, const ImageStruct *p_image)
 *		- returns a new texture with size, format and contents of p_image
 *		- p_state is OpenGL state guard
 *		- returns 0 on failure (bad pixel type, not enough memory)
 */
CGLTexture_2D *p_Create_ImageStruct_Texture(CGLState *p_state, const ImageStruct *p_image)
{
	int n_format;
	int n_data_type;
	int n_internal_format;

	if(!Get_GL_PixelType(p_image->PixelType, n_format, n_data_type, n_internal_format))
		return 0;
	// get pixel type

	int n_pixel_size = n_PixelSize(p_image->PixelType);
	// get pixel size to determine offsets in pixels instead of bytes

	if(abs(p_image->XOffset) == n_pixel_size && !(abs(p_image->YOffset) % n_pixel_size)) {
		void *p_raster = p_image->Raster;
		int n_y_offset;
		if(p_image->YOffset >= 0)
			n_y_offset = p_image->YOffset;
		else {
			n_y_offset = -p_image->YOffset;
			p_raster = (void*)((int)p_raster - n_y_offset * p_image->YSize);
		}
		int n_x_offset;
		if(p_image->XOffset >= 0)
			n_x_offset = p_image->XOffset;
		else {
			n_x_offset = -p_image->XOffset;
			p_raster = (void*)((int)p_raster - n_x_offset * p_image->XSize);
		}
		// find upper-left corner of image

		glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
		glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
		glPixelStorei(GL_UNPACK_ALIGNMENT, n_x_offset);
		glPixelStorei(GL_UNPACK_ROW_LENGTH, (n_y_offset != n_x_offset * p_image->XSize)?
			n_y_offset / n_pixel_size : 0);
		// set pixel transfer mode

		p_state->BindPixelUnPackBuffer(0);
		p_state->BindPixelPackBuffer(0);
		// explicitly bind no unpack buffer otherwise texture data would be read from there !!!

		CGLTexture_2D *p_texture;
		if(!(p_texture = new CGLTexture_2D(p_state, p_image->XSize, p_image->YSize,
		   n_internal_format, false, 0, n_format, n_data_type, p_raster)))
			return 0;
		if(!p_texture->b_Status()) {
			delete p_texture;
			return 0;
		}
		// create texture, copy image as a single block (fast)

		p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
		p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
		// clamp to edge

		return p_texture;
	}
	// directly create texture from image

	p_state->BindPixelUnPackBuffer(0);
	// explicitly bind no unpack buffer otherwise texture data would be read from there !!!

	CGLTexture_2D *p_texture;
	if(!(p_texture = new CGLTexture_2D(p_state, p_image->XSize, p_image->YSize,
	   n_internal_format, false)))
		return 0;
	if(!p_texture->b_Status()) {
		delete p_texture;
		return 0;
	}
	// create empty texture

	p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
	p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
	// clamp to edge

	CopyImage_To_Texture2D(p_image);
	// copy image to texture

	return p_texture;
}

/*
 *		-end-of-file-
 */
