/*
								+----------------------------------+
								|                                  |
								|     *** Resource manager ***     |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|            Manager.h             |
								|                                  |
								+----------------------------------+
*/

#ifndef __MANAGER_INCLUDED
#define __MANAGER_INCLUDED

/**
 *	template <class CResourceType>
 *	class CUnitCost
 *		- unit cost per each resource object
 */
template <class CResourceType>
class CUnitCost {
public:
	inline int n_Cost(const CResourceType &r_resource) const
	{
		return 1;
	}
};

/**
 *	template <class CResourceType>
 *	class CEvalCost
 *		- evaluate cost of each resource object by calling it's n_Size() const member function
 */
template <class CResourceType>
class CEvalCost {
public:
	inline int n_Cost(const CResourceType &r_resource) const
	{
		return return r_resource.n_Size();
	}
};

/**
 *	template <class CResourceType>
 *	class CFindIntId
 *		- function object for finding resource by it's id (std::find)
 */
template <class CResourceType>
class CFindIntId {
protected:
	int m_n_search_id;

public:
	inline CFindIntId(int n_search_id)
		:m_n_search_id(n_search_id)
	{}

	inline bool operator ()(const CResourceType &r_resource) const
	{
		return r_resource.n_Id() == m_n_search_id;
	}
};

/**
 *	class CStdStringCompare
 *		- class for comparing strings for std::strings
 */
class CStdStringCompare {
protected:
	const std::string &m_r_key;

public:
	inline CStdStringCompare(const std::string &r_key)
		:m_r_key(r_key)
	{}

	inline bool operator ()(const std::string &r_value) const
	{
		return !m_r_key.compare(r_value);
	}
};

/**
 *	template <class CResourceType>
 *	class CFindLastRecentlyUsed
 *		- function object for finding resource by it's last recently used
 */
template <class CResourceType>
class CFindLastRecentlyUsed {
protected:
	bool m_b_searching_first;
	unsigned int m_n_best_cost;

public:
	inline CFindLastRecentlyUsed()
		:m_b_searching_first(true)
	{}

	inline void Reset()
	{
		m_b_searching_first = true;
	}

	inline bool operator ()(const CResourceType &r_object, unsigned int n_last_used, int n_cost)
	{
		if(n_last_used < m_n_best_cost || m_b_searching_first) { // find objects with small n_last_used
			m_n_best_cost = n_last_used;
			return true;
		}
		return false;
	}
};

/**
 *	class CGenericObjectDestructor
 *		- class serving as generic CResourceDestructor in CResourceManager template
 */
class CGenericObjectDestructor {
public:
	template <class CObject>
	inline void operator ()(CObject *p_object)
	{
		_ASSERTE(p_object);
		delete p_object;
	}
};

/**
 *	template <class CResourceType, class CCostEvaluator = CEvalCost<CResourceType>,
 *		class TNameType = int, class CNameCompare = CFindIntId<CResourceType>,
 *		class TSourceType = std::string, class CSourceCompare = CStdStringCompare,
 *		class CResourceLoader, class CResourceDestructor>
 *	class CResourceManager
 *		- template class for universal resource manager
 *		- CResourceType is resource data type
 *		- CCostEvaluator is function object for evaluating resource costs
 *		- TNameType is object name type (usually integer)
 *		- CNameCompare is function object for comparing object names with reference value
 *		- TSourceType is data type to hold object source (i.e. constructor parameters,
 *		  address of datafile, etc ...)
 *		- CSourceCompare is function object for comparing (features CResourceType Load(TSourceType)
 *		  and static bool b_Valid_Object(CResourceType&) member functions)
 *		- CResourceDestructor is function object for freeing up objects
 */
template <class CResourceType, class CCostEvaluator = CEvalCost<CResourceType>,
	class TNameType = int, class CNameCompare = CFindIntId<CResourceType>,
	class TSourceType = std::string, class CSourceCompare = CStdStringCompare,
	class CResourceLoader, class CResourceDestructor = CGenericObjectDestructor>
class CResourceManager {
protected:
	struct TResourceWrap {
		unsigned int n_reference_num; // number of references (object with nonzero can't be freed)
		unsigned int n_recently_used; // contains time counter state when last referenced
		TSourceType t_source;
		CResourceType *p_object;
	};

	class CFindSource {
	protected:
		CSourceCompare m_compare;

	public:
		inline CFindSource(const TSourceType &r_key)
			:m_compare(r_key)
		{}

		inline bool operator ()(const TResourceWrap &r_t_wrap) const
		{
			return m_compare(r_t_wrap.t_source);
		}
	};

	class CFindUnusedSource {
	protected:
		CSourceCompare m_compare;

	public:
		inline CFindUnusedSource(const TSourceType &r_key)
			:m_compare(r_key)
		{}

		inline bool operator ()(TResourceWrap t_wrap)
		{
			return !t_wrap.n_reference_num && m_compare(t_wrap.t_source);
		}
	};

	class CFindNextSource {
	protected:
		CSourceCompare m_name_compare;
		CNameCompare m_prev_id_compare;
		bool m_b_already;

	public:
		inline CFindNextSource(const TNameType n_prev_id, const TSourceType &r_key)
			:m_name_compare(t_key), m_prev_id_compare(n_prev_id), m_b_already(false)
		{}

		inline bool operator ()(const TResourceWrap &r_t_wrap) const
		{
			if(m_b_already)
				return m_name_compare(r_t_wrap.t_source);
			else if(m_prev_id_compare(*r_t_wrap.p_object))
				m_b_already = true;
			return false;
		}
	};

	class CSumCost {
	protected:
		int m_n_cost;

	public:
		inline CSumCost()
			:m_n_cost(0)
		{}

		inline void operator ()(const TResourceWrap &r_t_wrap)
		{
			m_n_cost += CCostEvaluator().n_Cost(r_t_wrap.p_object);
		}

		inline operator int() const
		{
			return m_n_cost;
		}
	};

	template <class CFindBest>
	class CSelectPurge {
	protected:
		CFindBest &m_r_find_best;

	public:
		inline CSelectPurge(CFindBest &r_find_best)
			:m_r_find_best(r_find_best)
		{}

		inline bool operator ()(const TResourceWrap &r_t_wrap)
		{
			if(!r_t_wrap.n_reference_num) {
				return m_r_find_best(*r_t_wrap.p_object,
					r_t_wrap.n_recently_used, CCostEvaluator().n_Cost(*r_t_wrap.p_object));
				// supply CFindBest with last time object was used and resource cost
			}
			return false;
		}
	};

	std::vector<TResourceWrap> m_pool;
	unsigned int m_n_time;

	template <class CConsumer>
	class CUnwrapResource {
	protected:
		CConsumer &m_r_t_consumer;

	public:
		inline CUnwrapResource(CConsumer &r_t_consumer)
			:m_r_t_consumer(r_t_consumer)
		{}

		inline bool operator ()(TResourceWrap &r_t_wrap)
		{
			return m_r_t_consumer(*r_t_wrap.p_object);
		}

		inline operator CConsumer()
		{
			return m_r_t_consumer;
		}
	};

public:
	/**
	 *	CResourceManager()
	 *		- default constructor
	 */
	CResourceManager()
		:m_n_time(0)
	{}

	/**
	 *	~CResourceManager()
	 *		- default destructor
	 */
	~CResourceManager()
	{
		std::for_each(m_pool.begin(), m_pool.end(), FeedResourceDestructor);
	}

	void Clear()
	{
		std::for_each(m_pool.begin(), m_pool.end(), FeedResourceDestructor);
		m_pool.clear();
	}

	/**
	 *	bool InsertObject(CResourceType *p_object, const TSourceType t_source)
	 *		- insert object r_object with source t_source
	 *		- fail in case object with the same id is already contained in the pool
	 *		- returns true on success, false on failure
	 */
	bool InsertObject(CResourceType *p_object, const TSourceType t_source)
	{
		if(std::find_if(m_pool.begin(), m_pool.end(), CUnwrapResource<CNameCompare>(CNameCompare(p_object->n_Id()))) != m_pool.end())
			return false;
		// try to find object with the same id (then fail)

		m_pool.reserve(m_pool.size() + 1);
		if(m_pool.capacity() < m_pool.size() + 1)
			return false;
		// allocate enough capacity in the buffer

		TResourceWrap t_wrapper;
		t_wrapper.n_reference_num = 0;
		t_wrapper.n_recently_used = m_n_time;
		t_wrapper.t_source = t_source;
		t_wrapper.p_object = p_object;
		// fill wrapper structure

		m_pool.push_back(t_wrapper);
		// insert to the list

		return true;
	}

	/**
	 *	inline bool LoadObject(const TSourceType t_source, CResourceLoader &r_loader)
	 *		- load a new object from source t_source using r_loader
	 *		- fail in case object with the same id is already contained in the pool
	 *		- returns true on success, false on failure
	 */
	inline bool LoadObject(const TSourceType t_source, CResourceLoader &r_loader)
	{
		CResourceType *p_object = r_loader.Load(t_source);
		if(!p_object || !CResourceLoader::b_Valid_Object(p_object))
			return false;
		// load object ...

		if(!InsertObject(p_object, t_source)) {
			CResourceDestructor()(p_object);
			return false;
		}
		// ... and insert it, delete it in case it wasn't successful

		return true;
	}

	/**
	 *	bool DeleteObject(const TNameType n_id)
	 *		- deletes object with id n_id
	 *		- fails in case object was not found or in case it's reference count is not zero
	 *		- returns true on success, false on failure
	 */
	bool DeleteObject(const TNameType n_id)
	{
		typename std::vector<TResourceWrap>::iterator p_iter;
		if((p_iter = std::find_if(m_pool.begin(), m_pool.end(), CNameCompare(n_id))) == m_pool.end())
			return false;
		// try to find object with the same id (fail if not present)

		if((*p_iter).n_reference_num)
			return false;
		// can't delete object being in use

		CResourceDestructor()((*p_iter).p_object);
		// delete object

		m_pool.erase(p_iter, p_iter + 1);
		// remove from the list

		return true;
	}

	/**
	 *	const TNameType n_Object_Id(const TSourceType t_source, const TNameType n_bad_id) const
	 *		- returns id of first object in the pool whose source matches t_source
	 *		- in case no object was found, returns n_bad_id
	 */
	const TNameType n_Object_Id(const TSourceType t_source, const TNameType n_bad_id) const
	{
		typename std::vector<TResourceWrap>::const_iterator p_iter;
		if((p_iter = std::find_if(m_pool.begin(), m_pool.end(),
		   CFindSource(t_source))) == m_pool.end())
			return n_bad_id;
		// try to find object with the same source (fail if not present)

		return (*p_iter).p_object->n_Id();
		// return id
	}

	/**
	 *	const TNameType n_NextObject_Id(const TNameType n_prev_id,
	 *		const TSourceType t_source, const TNameType n_bad_id) const
	 *		- returns id of object, next to object with n_prev_id in the pool whose
	 *		  source matches t_source
	 *		- in case no object was found, returns n_bad_id
	 */
	const TNameType n_NextObject_Id(const TNameType n_prev_id,
		const TSourceType t_source, const TNameType n_bad_id) const
	{
		typename std::vector<TResourceWrap>::iterator p_iter;
		if((p_iter = std::find_if(m_pool.begin(), m_pool.end(),
		   CFindNextSource(n_prev_id, t_source))) == m_pool.end())
			return n_bad_id;
		// try to find next object with the same source (fail if not present)

		return (*p_iter).p_object->n_Id();
		// return id
	}

	/**
	 *	const TNameType n_UnusedObject_Id(const TSourceType t_source,
	 *		const TNameType n_bad_id) const
	 *		- returns id of object with zero references in the pool whose source matches t_source
	 *		- in case no object was found, returns n_bad_id
	 */
	const TNameType n_UnusedObject_Id(const TSourceType t_source, const TNameType n_bad_id) const
	{
		typename std::vector<TResourceWrap>::const_iterator p_iter;
		if((p_iter = std::find_if(m_pool.begin(), m_pool.end(),
		   CFindUnusedSource(t_source))) == m_pool.end())
			return n_bad_id;
		// try to find next object with the same source (fail if not present)

		return (*p_iter).p_object->n_Id();
		// return id
	}

	/**
	 *	inline unsigned int n_Object_Num() const
	 *		- return number of objects inside manager pool
	 */
	inline unsigned int n_Object_Num() const
	{
		return m_pool.size();
	}

	/**
	 *	bool IsObject(const TNameType n_id) const
	 *		- returns true in case n_id is id of some object, otherwise false
	 */
	bool IsObject(const TNameType n_id) const
	{
		return std::find_if(m_pool.begin(), m_pool.end(), CNameCompare(n_id)) != m_pool.end();
	}

	/**
	 *	int n_Reference_Num(const TNameType n_id) const
	 *		- returns reference count of object with id n_id
	 *		- in case there is no such object in the pool, returns -1
	 */
	int n_Reference_Num(const TNameType n_id) const
	{
		typename std::vector<TResourceWrap>::iterator p_iter;
		if((p_iter = std::find_if(m_pool.begin(), m_pool.end(), CNameCompare(n_id))) == m_pool.end())
			return -1;
		// try to find object with the same id (fail if not present)

		return (*p_iter).n_reference_num;
		// return reference count
	}

	/**
	 *	inline CResourceType &r_Object(const TNameType n_id)
	 *		- returns object with id n_id
	 *		- return value is undefined (invalid) in case no object with such an id exists
	 */
	inline CResourceType &r_Object(const TNameType n_id)
	{
		return *(*std::find_if(m_pool.begin(), m_pool.end(), CUnwrapResource<CNameCompare>(CNameCompare(n_id)))).p_object;
	}

	/**
	 *	inline const CResourceType &r_Object(const TNameType n_id) const
	 *		- returns object with id n_id
	 *		- return value is undefined (invalid) in case no object with such an id exists
	 */
	inline const CResourceType &r_Object(const TNameType n_id) const
	{
		return *(*std::find_if(m_pool.begin(), m_pool.end(), CUnwrapResource<CNameCompare>(CNameCompare(n_id)))).p_object;
	}

	/**
	 *	bool ReferenceObject(const TNameType n_id, bool b_allow_multiple_references)
	 *		- increases reference count of object with id n_id
	 *		- in case b_allow_multiple_references is false and given object has nonzero reference
	 *		  counter, function fails
	 *		- returns true in case such an object was found, otherwise false
	 */
	bool ReferenceObject(const TNameType n_id, bool b_allow_multiple_references)
	{
		typename std::vector<TResourceWrap>::iterator p_iter;
		if((p_iter = std::find_if(m_pool.begin(), m_pool.end(), CUnwrapResource<CNameCompare>(CNameCompare(n_id)))) == m_pool.end())
			return false;
		// try to find object with the same id (fail if not present)

		if(!b_allow_multiple_references && (*p_iter).n_reference_num)
			return false;
		// fail in case multiple references aren't allowed

		++ (*p_iter).n_reference_num;
		// increment counter
		
		(*p_iter).n_recently_used = m_n_time ++;
		// update recently used time, increment time counter

		return true;
	}

	/**
	 *	bool ReturnObject(const TNameType n_id, bool b_free_unused)
	 *		- decreases reference count of object with id n_id
	 *		- returns true in case such an object was found and it's reference
	 *		  count was not zero, otherwise false
	 */
	bool ReturnObject(const TNameType n_id, bool b_free_unused)
	{
		typename std::vector<TResourceWrap>::iterator p_iter;
		if((p_iter = std::find_if(m_pool.begin(), m_pool.end(), CUnwrapResource<CNameCompare>(CNameCompare(n_id)))) == m_pool.end())
			return false;
		// try to find object with the same id (fail if not present)

		if(!(-- (*p_iter).n_reference_num) && b_free_unused) {
			CResourceDestructor()((*p_iter).p_object);
			// delete object

			m_pool.erase(p_iter, p_iter + 1);
			// remove from the list
		}
		// decrease ref counter, free unused if required

		return true;
	}

	/**
	 *	int n_Cost_Sum() const
	 *		- returns sum of costs of all objects in the pool
	 */
	int n_Cost_Sum() const
	{
		return (int)std::for_each(m_pool.begin(), m_pool.end(), CSumCost());
	}

	/**
	 *	int n_Purge(int n_cost)
	 *		- deletes such amount of objects so their cost is equal or greater n_cost
	 *		- returns sum of costs of deleted objects (should be >= n_cost)
	 *		- note this frees objects in order they've been inserted to the pool
	 */
	int n_Purge(int n_cost)
	{
		int n_freed = 0;
		for(typename std::vector<TResourceWrap>::iterator p_iter = m_pool.begin();
		   n_freed < n_cost && p_iter != m_pool.end(); ++ p_iter) {
			if(!(*p_iter).n_reference_num) {
				n_freed += CCostEvaluator().n_Cost((*p_iter).p_object);
				// increase sum of data we already freed

				CResourceDestructor()((*p_iter).p_object);
				// delete object

				int n_index = p_iter - m_pool.begin();

				m_pool.erase(p_iter, p_iter + 1);
				// remove from the list

				p_iter = m_pool.begin() + (n_index - 1);
				// continue with the next iterator (pointers might change)
			}
		}
		return n_freed;
	}

	/**
	 *	template <class CFindBest>
	 *	int n_Purge_Selective(int n_cost, CFindBest t_find_best)
	 *		- deletes such amount of objects so their cost is equal or greater n_cost
	 *		- returns sum of costs of deleted objects (should be >= n_cost)
	 *		- note this uses CFindBest to find objects to be freed (ie slower than n_Purge_First())
	 *		  CFindBest can be CFindLastRecentlyUsed
	 */
	template <class CFindBest>
	int n_Purge_Selective(int n_cost, CFindBest t_find_best)
	{
		int n_freed = 0;
		while(n_freed < n_cost) {
			typename std::vector<TResourceWrap>::iterator p_iter, p_better;

			t_find_best.Reset();
			if((p_iter = std::find_if(m_pool.begin(), m_pool.end(),
			   CSelectPurge<CFindBest>(t_find_best))) == m_pool.end())
				break;
			// find first candidate

			while((p_better = std::find_if(p_iter + 1, m_pool.end(),
			   CSelectPurge<CFindBest>(t_find_best))) != m_pool.end())
				p_iter = p_better;
			// try to find a better one

			n_freed += CCostEvaluator().n_Cost(*(*p_iter).p_object);
			// increase sum of data we already freed

			CResourceDestructor()((*p_iter).p_object);
			// delete object

			m_pool.erase(p_iter, p_iter + 1);
			// remove from the list
		}
		return n_freed;
	}

protected:
	static inline void FeedResourceDestructor(TResourceWrap &r_t_wrap)
	{
		CResourceDestructor()(r_t_wrap.p_object);
	}
};

#endif //__MANAGER_INCLUDED
